/*!---------------------------------------------------------------------
  @file           ven962.cpp
  @author         JoergM
  @brief          vserver: NI support > 6.1 applications
  @see            example.html ...

\if EMIT_LICENCE
    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
\endif
---------------------------------------------------------------------*/


#define XDEBUG

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#if defined(OSF1) || defined(LINUX)
#include	<sys/wait.h>
#elif defined(SVR4)
#include	<wait.h>
#endif

#include "gsp01.h"
#include "hsp100.h"
#include "heo00.h"
#include "heo01.h"
#include "geo007_1.h"
#include "geo40.h"
#include "heo420.h"
#include "heo46.h"
#include "heo47.h"
#include "heo60.h"
#include "gen00.h"
#include "gen003.h"
#include "gen32.h"
#include "gen33.h"
#include "gen42.h"
#include "gen43.h"
#include "hen40.h"
#include "geo908.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
#define MOD__ "ven962.cpp: "

#define OLD_NISERVER_IDFILE "niserver.pid"

enum niserver_files { ni_protfile=0, ni_idfile=1, ni_tracefile=2, ni_oldidfile=3 };
static char const *ni_filesuffixes[] = { ".prot", ".pid", ".trace", "" };

static bool BuildNiServerFilename(niserver_files which, tsp00_Pathc &pathToFill, tsp01_RteError *RteError)
{
    if ( !sqlGetIndependentProtPath ( pathToFill, TERM_WITH_DELIMITER_EO01, RteError ) )
    {
        return false;
    }

    if ( ni_oldidfile == which )
    {
        if ( ( strlen(pathToFill.asCharp())
             + strlen(OLD_NISERVER_IDFILE)
             ) >= sizeof(tsp00_Pathc) )
        {
            eo46_set_rte_error ( RteError, undef_ceo00, "Path too long to file name:", pathToFill.asCharp() ) ;
            return false;
        }

        strcat ( pathToFill.asCharp() , OLD_NISERVER_IDFILE);

        return true;
    }

    static char myHostname[64] = { 0 };

    if ( !myHostname[0] )
    {
        if ( sql43_get_my_host_name(myHostname, sizeof(myHostname)) == -1 )
        {
            eo46_set_rte_error ( RteError, undef_ceo00, "Failed to get local maschine name", NULL ) ;
            return false;
        }
    }

    if ( ( strlen(pathToFill.asCharp())
         + strlen("niserver_")
         + strlen(myHostname)
         + strlen(ni_filesuffixes[(int)which])
         ) >= sizeof(tsp00_Pathc) )
    {
        eo46_set_rte_error ( RteError, undef_ceo00, "Path too long to file name:", pathToFill.asCharp() ) ;
        return false;
    }

    strcat ( pathToFill.asCharp() , "niserver_" );
    strcat ( pathToFill.asCharp() , myHostname);
    strcat ( pathToFill.asCharp() , ni_filesuffixes[(int)which]);

    return true;
}

#define ERRMSG_NISERVER_INUSE        "niserver already running!"

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/
typedef struct ten96_TCPIPConnectInfoRec
  {
  teo001_Pid                      WorkerPid ;
  teo00_Longuint                  ulConnStatus;
  teo00_Int4                      sd;
  void                           *NiHandle;
  tsp00_NodeIdc                   szSenderNodeID;

  teo00_Bool                      fAcknowledge;
  teo00_Longuint                  ulCserverStarted;
  teo00_Longuint                  ulSwapType;
  teo00_Longuint                  ulMaxSegmentSize;
  teo00_Longuint                  ulServiceType;
  teo00_Longuint                  ulServerRef;
  teo00_Longuint                  ulSenderRef;
  tsp00_DbNamec                   szServerDB;
  teo003_CommPacketRecord        *pCommPacket;
  teo00_Longuint                  ulMaxCmdDataLen;
  teo00_Longuint                  ulMinReplySize;
  connection_info                 Shm ;
  void *                          pDBM;
  } ten96_TCPIPConnectInfo;


class ServerID
{
private:

  	tsp00_Pathc      m_IDFile;
    teo00_Bool       m_IDFileCreated;
    teo00_Bool       m_IDFileExists;
    teo00_Bool       m_ServerIsRunning;
    pid_t            m_ServerPid;

public:
    ServerID                 (tsp01_RteError         *RteError);

    ~ServerID                () ;

    teo00_Bool CreateIDFile  (const pid_t             PidOfRunningServer,
                              tsp01_RteError         *RteError);

    teo00_Bool RemoveIDFile  (tsp01_RteError         *RteError);

    teo00_Bool KillServer    (tsp01_RteError         *RteError);

    teo00_Bool UnlistenServer(tsp01_RteError         *RteError);
};

//****************************************************************************
inline ServerID::~ServerID()
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("ServerID::~ServerID"));

  if ( m_IDFileCreated )
  {
     tsp01_RteError  RteError ;
     RemoveIDFile(&RteError);
  }
}

//****************************************************************************
inline teo00_Bool ServerID::CreateIDFile(const pid_t              PidOfRunningServer,
                                         tsp01_RteError          *RteError)
//****************************************************************************
{

  ROUTINE_DBG_MEO60 (_T("ServerID::CreateIDFile"));

  if ( m_IDFileExists && m_ServerIsRunning)
  {
    eo46_set_rte_error ( RteError, 0, ERRMSG_NISERVER_INUSE, NULL) ;
    DBG1 ((FUNCTION_NAME_MEO60,"%s exists and Server is running", (char *)m_IDFile )) ;
  }
  else
  {
    FILE *IdStream = fopen( m_IDFile, "wt");
    if ( IdStream )
    {
      DBG1 ((FUNCTION_NAME_MEO60,"writing pid %d to '%s'",PidOfRunningServer, (char *)m_IDFile )) ;
      if ( fprintf(IdStream, "%d", PidOfRunningServer ) > 0 )
      {
        fclose(IdStream);
      }
      else
      {
        eo46_set_rte_error ( RteError, undef_ceo00, "fprintf error", m_IDFile ) ;
        DBG1 ((FUNCTION_NAME_MEO60,"fprintf to %s:%s", (char *)m_IDFile, RteError->OsErrText ));
        fclose(IdStream);
      }
      m_IDFileCreated = true ;
      /* Allow only owner to read/write this file */
      chmod(m_IDFile, 0660);
    }
    else
    {
      eo46_set_rte_error ( RteError, undef_ceo00, "fopen error", m_IDFile ) ;
      DBG1 ((FUNCTION_NAME_MEO60,"fopen to %s:%s", (char *)m_IDFile, RteError->OsErrText ));
    }
  }

  return(RteError->RteErrCode == RTE_NO_ERROR_SP01);
}

/*---------------------------------------------------------------------------*/

//****************************************************************************
inline teo00_Bool ServerID::RemoveIDFile(tsp01_RteError          *RteError)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("ServerID::RemoveIDFile"));

  if ( m_IDFileExists || m_IDFileCreated )
    if ( unlink ( m_IDFile ) < 0 )
    {
      DBG1 ((FUNCTION_NAME_MEO60,"error unlink '%s':", (char *)m_IDFile, RteError->OsErrText));
      eo46_set_rte_error ( RteError, undef_ceo00, "unlink m_IDFile", m_IDFile ) ;
      return(RteError->RteErrCode == RTE_NO_ERROR_SP01);
    }
    else
    {
      m_IDFileCreated     = m_IDFileExists      = false ;
      m_ServerPid         = undef_ceo00 ;
    }

    return (true);
}

/*---------------------------------------------------------------------------*/

//****************************************************************************
inline ServerID::ServerID(tsp01_RteError          *RteError)
//****************************************************************************
{
  int oldIdFileInUse = 0;
  ROUTINE_DBG_MEO60 (_T("NIServerID::NIServerID"));

  // -- initialize private member
  m_ServerPid       = undef_ceo00 ;
  m_ServerIsRunning = false ;
  m_IDFileCreated   = false ;

  if ( BuildNiServerFilename( ni_oldidfile, m_IDFile, RteError ) )
  {
    // -- check if an old IdFile aready exists
    if ( (m_IDFileExists = (access ( m_IDFile, F_OK ) == 0)) )
    {
      FILE *IdStream ;
      DBG1 ((FUNCTION_NAME_MEO60,"m_IDFile '%s' exists", (char *)m_IDFile));

      if ( ( IdStream = fopen( m_IDFile, "rt" )) )
      {
        char PidBuf[10];
        if ( fgets(PidBuf, sizeof(PidBuf)-1, IdStream ) > 0 )
        {
          DBG1 ((FUNCTION_NAME_MEO60,"PidBuf '%s' ", PidBuf ));
          m_ServerPid       = atoi ( PidBuf ) ;
          m_ServerIsRunning = ( m_ServerPid > 1 ) && ( kill(m_ServerPid,0) == 0 || errno != ESRCH );
          DBG1 ((FUNCTION_NAME_MEO60,"m_ServerPid '%d', m_ServerIsRunning %s",
                  m_ServerPid, m_ServerIsRunning ? "YES" : "NO"));
        }
        else
        {
          eo46_set_rte_error ( RteError, undef_ceo00, "fgets error", m_IDFile ) ;
          DBG1 ((FUNCTION_NAME_MEO60,"error fgets '%s': %s", m_IDFile, RteError->OsErrText));
        }

        fclose(IdStream);

        if ( !m_ServerIsRunning )
        {
          DBG1 ((FUNCTION_NAME_MEO60,"removing '%s' because server is not running", (char *)m_IDFile));
          RemoveIDFile(RteError);
        }
      }
      else
      {
        eo46_set_rte_error ( RteError, undef_ceo00, "fopen error", m_IDFile ) ;
        DBG1 ((FUNCTION_NAME_MEO60,"error fopen '%s': %s", (char *)m_IDFile, RteError->OsErrText));
      }
    }
  }

  if ( !m_ServerIsRunning && BuildNiServerFilename( ni_idfile, m_IDFile, RteError ) )
  {
    // -- check if m_IDFile aready exists
    if ( (m_IDFileExists = (access ( m_IDFile, F_OK ) == 0)) )
    {
      FILE *IdStream ;
      DBG1 ((FUNCTION_NAME_MEO60,"m_IDFile '%s' exists", (char *)m_IDFile));

      if ( ( IdStream = fopen( m_IDFile, "rt" )) )
      {
        char PidBuf[10];
        if ( fgets(PidBuf, sizeof(PidBuf)-1, IdStream ) > 0 )
        {
          DBG1 ((FUNCTION_NAME_MEO60,"PidBuf '%s' ", PidBuf ));
          m_ServerPid       = atoi ( PidBuf ) ;
          m_ServerIsRunning = ( m_ServerPid > 1 ) && ( kill(m_ServerPid,0) == 0 || errno != ESRCH );
          DBG1 ((FUNCTION_NAME_MEO60,"m_ServerPid '%d', m_ServerIsRunning %s",
                  m_ServerPid, m_ServerIsRunning ? "YES" : "NO"));
        }
        else
        {
          eo46_set_rte_error ( RteError, undef_ceo00, "fgets error", m_IDFile ) ;
          DBG1 ((FUNCTION_NAME_MEO60,"error fgets '%s': %s", m_IDFile, RteError->OsErrText));
        }

        fclose(IdStream);

        if ( !m_ServerIsRunning )
        {
          DBG1 ((FUNCTION_NAME_MEO60,"removing '%s' because server is not running", (char *)m_IDFile));
          RemoveIDFile(RteError);
        }
      }
      else
      {
        eo46_set_rte_error ( RteError, undef_ceo00, "fopen error", m_IDFile ) ;
        DBG1 ((FUNCTION_NAME_MEO60,"error fopen '%s': %s", (char *)m_IDFile, RteError->OsErrText));
      }
    }
  }
}


//****************************************************************************
inline teo00_Bool ServerID::KillServer(tsp01_RteError          *RteError)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("ServerID::KillServer"));

  if ( m_ServerIsRunning && (m_ServerPid > 1) )
  {
    DBG1 ((FUNCTION_NAME_MEO60,"try to kill %d", -m_ServerPid ));
    if ( kill ( (pid_t) (- m_ServerPid) , SIGTERM ) != 0 )
    {
      DBG1 ((FUNCTION_NAME_MEO60,"try to kill %d", m_ServerPid ));
      if ( kill ( m_ServerPid , SIGTERM ) != 0 )
      {
        eo46_set_rte_error ( RteError, undef_ceo00, "kill server", NULL ) ;
      }
    }
  }

  if ( RteError->RteErrCode == RTE_NO_ERROR_SP01)
  {
    m_ServerIsRunning = false ;
    RemoveIDFile(RteError);
  }

  return(RteError->RteErrCode == RTE_NO_ERROR_SP01);
}

//****************************************************************************
inline teo00_Bool ServerID::UnlistenServer(tsp01_RteError          *RteError)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("ServerID::UnlistenServer"));

  RteError->RteErrCode = RTE_NO_ERROR_SP01;

  if ( m_ServerIsRunning && (m_ServerPid > 1) )
  {
    DBG1 ((FUNCTION_NAME_MEO60,"try to kill %d", m_ServerPid ));
    if ( kill ( (pid_t) m_ServerPid, SIGKILL ) != 0 )
    {
        eo46_set_rte_error ( RteError, undef_ceo00, "kill server", NULL ) ;
    }
  }

  if ( RteError->RteErrCode == RTE_NO_ERROR_SP01)
  {
    m_ServerIsRunning = false ;
    RemoveIDFile(RteError);
  }

  return(RteError->RteErrCode == RTE_NO_ERROR_SP01);
}


/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

static  int       sql03_timeout ;
extern	char     *sql22_file ;
extern	int       sql22_timestamp ;
extern  int       sql22_fd;
extern  int       sql22_keep_open;



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

static  ten96_TCPIPConnectInfo         *GlobalConnnectInfo = NULL;
/* local data */

static  int                             e96_enable_timestamp ;
static  int                             e96_interactive ;
static  pid_t                           e96_master_pid ;
static  pid_t                           e96_pid ;
static  struct sockaddr                 e96_sa ;
static  unsigned short                  en962_NiServicePort = 0 ; /* NI default (3299) */
static  void                           *en962_NiListenHandle ;

static  connection_info *e96_cip ;

static	SQL_DBNAMEC		                  e96_remote_db ;
static	tsp00_Int4		                    e96_remote_ref ;
static	int			                        e96_remote_swap ;

static  tsp00_Versionc   en962_RTEVersionString;

/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/
extern "C" {
    typedef void (*SignalHandler)(int);
};
static  SIGNALFUNCTYPE  e96_catch_sigcld     ( int           sig);
static  SIGNALFUNCTYPE  e96_catch_signal     ( int			     sig ) ;
static  int             e96_SetSigAction     ( int           sig,
                                               SignalHandler Handler );
static  void            en960_SetStdSignals  ( int           IsInteractive);

static  teo00_Longuint	
                   en962InfoRequestToDB ( ten96_TCPIPConnectInfo        *pConnInfo,
                                          teo003_ConPktParamRecord      *pConPktParamRec,
                                          tsp00_ErrTextc                 pErrText );

static  teo00_Bool en962_AnalyzeParamAndSetGlobals (int              argc ,
                                                    char           **argv,
                                                    teo00_Bool      *killServer,
                                                    teo00_Bool      *unlistenServer,
                                                    tsp01_RteError  *RteError);

static  void       en962_CloseHandles () ;

static  void       en962_SetProtocolFile ( tsp01_RteError  *RteError );

static  teo00_Bool en962_InitNi (unsigned short  *NiServicePort,
                                 void           **NiListenHandle,
                                 tsp01_RteError  *RteError );

static  teo00_Bool en962_NiCoordinator (int              StartupPipe[2],
                                        ServerID        &ServerID,
                                        tsp01_RteError  *RteError);

static  teo00_Bool en962_SetPidAndProcessGroup (pid_t           *MyPid ,
                                                pid_t           *MasterPid ,
                                                tsp01_RteError  *RteError ) ;

static  teo00_Bool en962_NiCoordinatorInit (int              StartupPipe[2],
                                            ServerID        &ServerID,
                                            unsigned short  *NiServicePort,
                                            void           **NiListenHandle,
                                            tsp01_RteError  *RteError);

static  void       en962_NiWorker          (ten96_TCPIPConnectInfo *pConnInfo,
                                            tsp01_RteError         *RteError);

static void
en962_InitShmStructFromConnPktParam ( ten96_TCPIPConnectInfo   *pConnInfo,
                                      teo003_ConPktParamRecord *pConPktParamRec) ;

static teo00_Longuint
en962_NiGetRequest (ten96_TCPIPConnectInfo *pConnInfo) ;

static teo00_Longuint
en962_NiTransfer   (ten96_TCPIPConnectInfo *pConnInfo) ;

static teo00_Longuint
en962_NiServerConnectToDBMSrv (ten96_TCPIPConnectInfo   *pConnInfo,
                               teo003_ConPktParamRecord *pConPktParamRec );

static teo00_Longuint	
en962_NiServerInfoRequestToDB ( ten96_TCPIPConnectInfo        *pConnInfo,
                                teo003_ConPktParamRecord      *pConPktParamRec,
                                tsp00_ErrTextc                 pErrText );


static teo00_Longuint
en962_NiServerConnectToDB (ten96_TCPIPConnectInfo   *pConnInfo,
                           teo003_ConPktParamRecord *pConPktParamRec );

static teo00_Longuint
en962_NiServerCancelOrDumpSession (ten96_TCPIPConnectInfo   *pConnInfo,
                                   teo003_ConPktParamRecord *pConPktParamRec ) ;

/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

//****************************************************************************
int     main ( int argc, char  **argv )
//****************************************************************************
{
    ROUTINE_DBG_MEO60 (_T("main"));
    int                     rc ;
    int                     StartupPipe [ 2 ];
    tsp01_RteError          RteError;
    teo00_Bool              killServer;
    teo00_Bool              unlistenServer;

    /* Server input may be piped,               */
    /* so close stdin to free the process(es)   */
    /* that write to the pipe.                  */
    /* Reopen stdin, so file descriptors 0      */
    /* will not be used for network operations. */
    /* The above is important for both the      */
    /* starter and the background server.       */

    /* reenable redirection of output. This coding inform 'MSGD' output to write on stderr and keep it open */
    sql22_fd = 2;
    sql22_keep_open = 1;

    sp100_GetVersionString((char *)COMP_NAME_RTE_SP100, s100buildnumber, en962_RTEVersionString);
    eo46_rte_error_init (&RteError);

    ServerID  NIServerID(&RteError) ;

    en962_CloseHandles();

    if ( !en962_AnalyzeParamAndSetGlobals (argc, argv, &killServer, &unlistenServer, &RteError ) )
      return(1);

    if ( killServer )
    {
      NIServerID.KillServer (&RteError);
      return(0);
    }

    if ( unlistenServer )
    {
      NIServerID.UnlistenServer (&RteError);
      return(0);
    }

    DBG1 ((FUNCTION_NAME_MEO60,"starting \n" ));

    /*
     *  File permission mode is always specified absolutely
     *  within the server.
     */
    (void) umask ( 007 );

    en960_SetStdSignals ( e96_interactive );

	rc = pipe ( StartupPipe );
	if ( rc )
	{
        eo46_set_rte_error ( &RteError, undef_ceo00, "create pipe", NULL ) ;
	    DBG1 ((FUNCTION_NAME_MEO60," :cannot create pipe, %s\n" , RteError.OsErrText )) ;
	    return ( rc );
	}
	/* create background server */
    switch (fork ())
    {
    case  0 : sql22_fd = StartupPipe[1];
              sql22_keep_open = 1;
              en962_NiCoordinator (StartupPipe, NIServerID, &RteError);
              break ;

    case -1 : eo46_set_rte_error ( &RteError, undef_ceo00, "error during fork", NULL ) ;
	          MSGD (( ERR_NI_SERV_FORK , RteError.OsErrText ));
              rc = -1 ;
              break ;

    default : close ( StartupPipe [ 1 ] ); /* don't write to pipe */
              /* starter exits when pipe becomes closed */
	          for ( ; ; )
              {
                char                    line [ 256 ];
                int                     lgt ;

		        lgt = read ( StartupPipe [ 0 ] , line , sizeof(line) );
		        if ( lgt <= 0 ) break ;
		        write ( 2 , line , (unsigned) lgt );
              }
              rc = 0 ;
    }
  return(rc);
}


/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/


//****************************************************************************
static  teo00_Bool en962_NiCoordinator (int              StartupPipe[2],
                                        ServerID        &ServerID,
                                        tsp01_RteError  *RteError)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_NiCoordinator"));

  ten96_TCPIPConnectInfo  ConnInfo ;
  teo00_Longuint          CommState = commErrOk_esp01;
  tsp00_ErrTextc          pErrText ;

  GlobalConnnectInfo    = &ConnInfo ;

  memset( &ConnInfo, 0, sizeof(ConnInfo));
  en962_NiCoordinatorInit (StartupPipe, ServerID,
                           &en962_NiServicePort, &en962_NiListenHandle, RteError) ;

  if ( RteError->RteErrCode == RTE_NO_ERROR_SP01 )
  {
    do
    {
      DBG1 ((FUNCTION_NAME_MEO60,"waiting for connection ..." ));
      CommState = eo40NiAccept ( en962_NiListenHandle, &ConnInfo.NiHandle,
                                 ConnInfo.szSenderNodeID, sizeof(ConnInfo.szSenderNodeID),
                                 pErrText);

      if ( CommState == commErrOk_esp01 )
      {
    	DBG1 ((FUNCTION_NAME_MEO60,"accepted connection sd %d \n" , ConnInfo.NiHandle ));
        switch (fork ())
        {
        case  0 : eo40NiClose ( &en962_NiListenHandle );
                  en962_NiWorker (&ConnInfo, RteError);
                  break ;

        case -1 : eo46_set_rte_error ( RteError, undef_ceo00, "error during fork", NULL ) ;
                  MSGD (( ERR_NI_SERV_FORK , RteError->OsErrText ));
                  break ;

        default : eo40NiClose ( &ConnInfo.NiHandle  );
        }
      }
      else
      {
        MSGD (( ERR_NI_SERV_ACCEPT_ERROR, (char *)pErrText ));
      }
    }
    while ( CommState == commErrOk_esp01 && RteError->RteErrCode == RTE_NO_ERROR_SP01) ;
    eo40NiFinish(pErrText);

    MSGD (( INFO_NI_SERV_STOPPED ))
  }

  return ( CommState == commErrOk_esp01 && RteError->RteErrCode == RTE_NO_ERROR_SP01 );
}

//****************************************************************************
static  teo00_Bool en962_NiCoordinatorInit (int              StartupPipe[2],
                                            ServerID        &ServerID,
                                            unsigned short  *NiServicePort,
                                            void           **NiListenHandle,
                                            tsp01_RteError  *RteError)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_NiCoordinatorInit"));


  /* background server */

  (void) close ( StartupPipe[0] ); /* don't read from pipe */
  StartupPipe[0] = -1 ;

  /* Server output may be piped,                       */
  /* so close stdout/stderr to free the process(es)    */
  /* that read from the pipe(s).                       */
  /* Reopen stdout/stderr, so file descriptors 1 and 2 */
  /* will not be used for network operations.          */

  /* redirect stdout to the pipe */
  (void) close ( 1 );
  (void) dup ( StartupPipe[1] );

  /* redirect stderr to the pipe */
  (void) close ( 2 );
  (void) dup ( StartupPipe[1] );

  if ( !en962_SetPidAndProcessGroup ( &e96_pid, &e96_master_pid, RteError ) ||
       !ServerID.CreateIDFile       ( e96_pid, RteError ) )
  {
    MSGD (( ERR_NI_SERV_DURING_STARTUP , RteError->RteErrText ));
    if ( RteError->OsErrCode )
    {
      MSGD (( ERR_NI_SERV_DURING_STARTUP , RteError->OsErrText ));
    }
  }
  else
  {
    if (  !en962_InitNi(NiServicePort, NiListenHandle, RteError) )
    {
      MSGD (( ERR_NI_SERV_DURING_STARTUP , RteError->RteErrText ))
    }
    else
    {
      MSGD (( INFO_NI_SERV_STARTED ))
      /* from now on, debug output will have a time stamp */
      en962_SetProtocolFile ( RteError );
    }
  }

  /* closing the pipe lets the starter exit */
  (void) close ( StartupPipe[1] );
  StartupPipe[1] = -1 ;

  return ( RteError->RteErrCode == RTE_NO_ERROR_SP01 );
}

//****************************************************************************
static  void
en962_NiWorker          (ten96_TCPIPConnectInfo *pConnInfo,
                         tsp01_RteError         *RteError)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_NiWorker"));

  teo00_Longuint                  ulCommState;
  tsp00_ErrTextc                  pErrText ;

  DBG1 ((FUNCTION_NAME_MEO60, INFO_NI_SERV_STARTED ));

  pConnInfo->WorkerPid    = getpid();
  pConnInfo->ulConnStatus = CON_INITIALIZED_EO003;

  do
  {
    ulCommState = en962_NiGetRequest(pConnInfo);
  }
  while ( ulCommState             == commErrOk_esp01 &&
          pConnInfo->ulConnStatus == CON_KEEP_ALIVE_EO003 );

  if ( pConnInfo->ulConnStatus == CON_CONNECTED_EO003 )
  {
    if ( ulCommState == commErrOk_esp01 )
    {
      DBG1 ((FUNCTION_NAME_MEO60, INFO_XSER_APPL_CONNECTED,
              (char *)pConnInfo->szSenderNodeID,
              (char *)pConnInfo->szServerDB,
              pConnInfo->ulServerRef ));

      ulCommState = en962_NiTransfer ( pConnInfo );

      DBG1 ((FUNCTION_NAME_MEO60, INFO_XSER_APPL_DISCONNECT,
              (char *)pConnInfo->szSenderNodeID,
              (char *)pConnInfo->szServerDB,
              pConnInfo->ulServerRef ));
    }

    if ( pConnInfo->pDBM != NULL )
      {
      eo908_Release ( pConnInfo->pDBM );
      pConnInfo->pDBM = NULL;
      }
    else
      (void) sql33_release (  &pConnInfo->Shm, pErrText, 0, 0 );

    pConnInfo->ulConnStatus = CON_RELEASED_EO003;
  }


  eo40NiClose ( &pConnInfo->NiHandle );
  eo40NiFinish(pErrText);

  DBG1 ((FUNCTION_NAME_MEO60, INFO_NI_SERV_STOPPED ))
  exit( ulCommState );
}

//****************************************************************************
static void
en962_InitShmStructFromConnPktParam( ten96_TCPIPConnectInfo   *pConnInfo,
                                     teo003_ConPktParamRecord *pConPktParamRec)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_InitShmStructFromConnPktParam"));

  pConnInfo->Shm.ci_mess_class       = pConPktParamRec->ulMessClass ;
  pConnInfo->Shm.ci_service          = pConPktParamRec->ulServiceType ;
  pConnInfo->Shm.ci_max_segment_size = pConPktParamRec->ulMaxSegmentSize ;
  pConnInfo->Shm.ci_max_data_size    = pConPktParamRec->ulMaxDataLen ;
  pConnInfo->Shm.ci_packet_size      = pConPktParamRec->ulPacketSize ;
  pConnInfo->Shm.ci_min_reply_size   = pConPktParamRec->ulMinReplySize ;
  pConnInfo->Shm.ci_my_pid           = pConnInfo->WorkerPid ;

  pConnInfo->Shm.ci_packet_cnt = 1 ;
  pConnInfo->Shm.ci_my_ref     = 1 ;
  pConnInfo->Shm.ci_connect_time     = time ( (time_t*)0 );
  pConnInfo->Shm.ci_protocol   = RSQL_RTE_PROT_LOCAL_EO003 ;
  pConnInfo->Shm.ci_state      = CON_CONNECTING ;
  pConnInfo->Shm.ci_use_count++ ;

  sql47_ltoa(pConPktParamRec->pidSenderPID, (char *)pConnInfo->Shm.ci_remote_pid,
             sizeof(pConnInfo->Shm.ci_remote_pid));

  strcpy( pConnInfo->Shm.ci_peer_node  , pConnInfo->szSenderNodeID );
  strcpy( pConnInfo->Shm.ci_peer_dbname, pConPktParamRec->szReceiverServerDB ) ;

}

//****************************************************************************
static teo00_Longuint	
en962_NiServerInfoRequestToDB ( ten96_TCPIPConnectInfo   *pConnInfo,
                                teo003_ConPktParamRecord *pConPktParamRec,
                                tsp00_ErrTextc            pErrText )
//****************************************************************************
{
    ROUTINE_DBG_MEO60 (_T("en962_NiServerInfoRequestToDB"));
    teo00_Longuint         ulCommState;

    (void) alarm ( LOCL_CON_TMO );

    DBG1 ((FUNCTION_NAME_MEO60, "Received info request, SERVERDB: %s",
          (char *)ConPktParamRec.szReceiverServerDB ));

    ulCommState = en962InfoRequestToDB( pConnInfo, pConPktParamRec, pErrText );

    pConnInfo->ulConnStatus = CON_KEEP_ALIVE_EO003;

    strcpy ( pConPktParamRec->szSenderServerDB,
             pConPktParamRec->szReceiverServerDB );
    pConPktParamRec->szReceiverServerDB[0] = '\0';
    pConPktParamRec->ulReceiverRef         = pConPktParamRec->ulSenderRef;
    pConPktParamRec->ulSenderRef           = (ULONG)undef_ceo00;
    pConPktParamRec->usServicePort         = (tsp00_Uint2)undef_ceo00; /* unused */
    pConPktParamRec->ulMessClass           = RSQL_INFO_REPLY_EO003;
    pConPktParamRec->ulCommState           = ulCommState;

    (void) alarm ( 0 );

    return ulCommState;
}

//****************************************************************************
static teo00_Longuint
en962_NiServerConnectToDB (ten96_TCPIPConnectInfo   *pConnInfo,
                           teo003_ConPktParamRecord *pConPktParamRec )
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_NiServerConnectToDB"));
  teo00_Longuint         ulCommState;
  tsp00_ErrTextc         pErrText;

  DBG1 ((FUNCTION_NAME_MEO60,
         "Connect req. received, SERVERDB: %s, SenderRef: %d, PacketSize: %d, MaxDataLen: %d",
          (char *)pConPktParamRec->szReceiverServerDB,
          pConPktParamRec->ulSenderRef, pConPktParamRec->ulPacketSize,
          pConPktParamRec->ulMaxDataLen ));


  (void) alarm ( LOCL_CON_TMO );

  ulCommState = sql33_connect ( &pConnInfo->Shm, pErrText,
                               (struct connection_info**)0, 0 );

  DBG1 ((FUNCTION_NAME_MEO60, "sql33_connect returned %d, ErrMsg=%s", ulCommState, pErrText.asCharp()));

  if ( ulCommState == commErrOk_esp01 )
  {
    pConnInfo->ulConnStatus   = CON_CONNECTED;
    pConnInfo->ulSwapType     = pConPktParamRec->ulSwapType ;
    pConnInfo->Shm.ci_request = pConnInfo->Shm.ci_packet[0];
    pConnInfo->pCommPacket    = REINTERPRET_CAST(teo003_CommPacketRecord *,pConnInfo->Shm.ci_request);

    pConnInfo->ulMinReplySize   = pConnInfo->Shm.ci_min_reply_size;
    pConnInfo->ulMaxCmdDataLen  = pConnInfo->Shm.ci_max_data_size  -
                                  pConnInfo->Shm.ci_min_reply_size;

    pConnInfo->ulMaxSegmentSize = pConnInfo->Shm.ci_max_segment_size ;
    pConnInfo->ulServiceType    = pConPktParamRec->ulServiceType;

    DBG1 ((FUNCTION_NAME_MEO60,"Connect req. success, MaxCmdDataLen: %d, MaxSegmentSize: %d",
            pConnInfo->ulMaxCmdDataLen, pConnInfo->ulMaxSegmentSize )) ;
  }
  strcpy ( pConPktParamRec->szSenderServerDB,
           pConPktParamRec->szReceiverServerDB );

  pConPktParamRec->szReceiverServerDB[0] = '\0';
  pConPktParamRec->ulReceiverRef  = pConPktParamRec->ulSenderRef;
  pConPktParamRec->ulSenderRef    = pConnInfo->Shm.ci_peer_ref;
  pConPktParamRec->ulPacketSize   = pConnInfo->Shm.ci_packet_size;
  pConPktParamRec->ulMaxDataLen   = pConnInfo->Shm.ci_max_data_size;
  pConPktParamRec->ulMinReplySize = pConnInfo->Shm.ci_min_reply_size;
  pConPktParamRec->ulMessClass    = RSQL_USER_CONN_REPLY_EO003;

  (void) alarm ( 0 );

  return ( ulCommState );
}

//****************************************************************************
static teo00_Longuint
en962_NiServerConnectToDBMSrv (ten96_TCPIPConnectInfo   *pConnInfo,
                               teo003_ConPktParamRecord *pConPktParamRec )
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_NiServerConnectToDBMSrv"));
  teo00_Longuint        ulCommState;
  tsp00_Uint4           TransferBufferSize;

  DBG1 ((FUNCTION_NAME_MEO60,
         "Connect req. received, SERVERDB: %s, SenderRef: %d, PacketSize: %d, MaxDataLen: %d",
          (char *)pConPktParamRec->szReceiverServerDB,
          pConPktParamRec->ulSenderRef, pConPktParamRec->ulPacketSize,
          pConPktParamRec->ulMaxDataLen ));


  (void) alarm ( LOCL_CON_TMO );

  tsp00_Uint4 MaxCmdDataLen;
  tsp00_Uint4 ServerRef;

  ulCommState = eo908_Connect( "dbmsrv",
                               pConPktParamRec->szServerDBRoot,
                               pConPktParamRec->szReceiverServerDB,
                               &pConnInfo->pDBM,
                               (void**)&pConnInfo->pCommPacket,
                               &MaxCmdDataLen,
                               &TransferBufferSize,
                               &ServerRef );

  DBG1 ((FUNCTION_NAME_MEO60, "eo908_Connect returned %d", ulCommState));

  if ( ulCommState == commErrOk_esp01 )
  {
    pConnInfo->ulMaxCmdDataLen  = MaxCmdDataLen;
    pConnInfo->ulServerRef      = ServerRef;
    pConnInfo->ulSenderRef      = pConPktParamRec->ulSenderRef;
    pConnInfo->ulMaxSegmentSize = pConPktParamRec->ulMaxSegmentSize;
    pConnInfo->fAcknowledge     = pConPktParamRec->fAcknowledge;
    pConnInfo->ulSwapType       = pConPktParamRec->ulSwapType;
    pConnInfo->ulServiceType    = pConPktParamRec->ulServiceType;
    pConnInfo->ulConnStatus     = CON_CONNECTED;

    DBG1 ((FUNCTION_NAME_MEO60,"Connect req. success, MaxCmdDataLen: %d, MaxSegmentSize: %d",
            pConnInfo->ulMaxCmdDataLen, pConnInfo->ulMaxSegmentSize )) ;
  }

  strcpy ( pConPktParamRec->szSenderServerDB,
           pConPktParamRec->szReceiverServerDB );

  pConPktParamRec->szReceiverServerDB[0] = '\0';
  pConPktParamRec->ulReceiverRef  = pConPktParamRec->ulSenderRef;
  pConPktParamRec->ulSenderRef    = pConnInfo->ulServerRef;
  pConPktParamRec->ulMessClass    = RSQL_CTRL_CONN_REPLY_EO003;

  (void) alarm ( 0 );

  return ( ulCommState );
}

//****************************************************************************
static teo00_Longuint
en962_NiServerCancelOrDumpSession (ten96_TCPIPConnectInfo   *pConnInfo,
                                   teo003_ConPktParamRecord *pConPktParamRec )
{
  ROUTINE_DBG_MEO60 (_T("en962_NiServerCancelOrDumpSession"));
  teo00_Longuint         ulCommState;
  tsp00_ErrTextc         pErrText;

  switch ( pConPktParamRec->ulMessClass )
  {
  case RSQL_CTRL_CANCEL_REQUEST_EO003 :
    DBG1 ((FUNCTION_NAME_MEO60, "Cancel request, DBMSERVER reference: %d", pConPktParamRec->ulReceiverRef  ));
    pConnInfo->ulConnStatus = CON_CANCELLED_EO003;
    break ;

  case RSQL_USER_CANCEL_REQUEST_EO003 :

    DBG1 ((FUNCTION_NAME_MEO60, "Cancel request, server reference: %d", pConPktParamRec->ulReceiverRef  ));
    pConnInfo->Shm.ci_peer_ref = pConPktParamRec->ulReceiverRef ;
    ulCommState = sql33_cancel ( &pConnInfo->Shm, pErrText);
    pConnInfo->ulConnStatus = CON_CANCELLED_EO003;
    break ;

  case RSQL_DUMP_REQUEST_EO003:

    DBG1 ((FUNCTION_NAME_MEO60, "Dump request, server reference: %d", pConPktParamRec->ulReceiverRef  ));
    ulCommState = sql33_dump ( &pConnInfo->Shm , pErrText );
    pConnInfo->ulConnStatus = CON_CANCELLED_EO003;
    break ;

  default : ulCommState = commErrNotOk_esp01 ;
  }

  return ( ulCommState );
}

//****************************************************************************
static teo00_Longuint
en962_NiGetRequest (ten96_TCPIPConnectInfo *pConnInfo)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_NiGetRequest"));

  teo00_Longuint                  ulCommState;
  teo003_ConPktParamRecord        ConPktParamRec;
  tsp00_ErrTextc                  pErrText ;

  ulCommState = eo420ReceiveConnectPacket ( pConnInfo->NiHandle,
                                            (RECV_PKT_FUNC_PTR_EO003)eo40NiReceive,
                                            &ConPktParamRec, pErrText );

  DBG1 ((FUNCTION_NAME_MEO60,
          "Rec. request packet - Class. %d, MyRef: %d, PeerRef: %d, Service: %d, SERVERDB '%s', ClientDB: '%s'",
          ConPktParamRec.ulMessClass,
          ConPktParamRec.ulReceiverRef,
          ConPktParamRec.ulSenderRef,
          ConPktParamRec.ulServiceType,
          (char *)ConPktParamRec.szReceiverServerDB,
          (char *)ConPktParamRec.szSenderServerDB ));

  en962_InitShmStructFromConnPktParam(pConnInfo, &ConPktParamRec ) ;

  switch ( ConPktParamRec.ulMessClass )
    {
    /****************************/
    case RSQL_INFO_REQUEST_EO003:
    case RSQL_INFO_REQUEST_KEEP_ALIVE_EO003:
    /****************************/

      ulCommState = en962_NiServerInfoRequestToDB ( pConnInfo, &ConPktParamRec, pErrText );

      break;

    /****************************/
    case RSQL_USER_CONN_REQUEST_EO003 :
    /****************************/

      ulCommState = en962_NiServerConnectToDB ( pConnInfo, &ConPktParamRec ) ;

      if ( ulCommState == commErrOk_esp01 )
      {
        ulCommState = eo420SetNiSocketBufferSize ( pConnInfo->NiHandle, 1024,
                                                  &ConPktParamRec.ulMaxSegmentSize,
                                                   pErrText) ;
        pConnInfo->Shm.ci_max_segment_size = pConnInfo->ulMaxSegmentSize =
                                             ConPktParamRec.ulMaxSegmentSize;
      }
      break;

    /****************************/
    case RSQL_CTRL_CONN_REQUEST_EO003:
    /****************************/

      ulCommState = en962_NiServerConnectToDBMSrv ( pConnInfo, &ConPktParamRec ) ;

      if ( ulCommState == commErrOk_esp01 )
      {
        ulCommState = eo420SetNiSocketBufferSize ( pConnInfo->NiHandle, 1024,
                                                  &ConPktParamRec.ulMaxSegmentSize,
                                                   pErrText) ;
        pConnInfo->Shm.ci_max_segment_size = pConnInfo->ulMaxSegmentSize =
                                             ConPktParamRec.ulMaxSegmentSize;
      }


      break;

    /****************************/
    case RSQL_USER_CANCEL_REQUEST_EO003:
    case RSQL_DUMP_REQUEST_EO003       :
    case RSQL_CTRL_CANCEL_REQUEST_EO003:
    /****************************/

      ulCommState = en962_NiServerCancelOrDumpSession (pConnInfo, &ConPktParamRec ) ;
      return ( ulCommState );


    /* PTS 1105295 */
    /****************************/
    case RSQL_PING_REQUEST_EO003:
    /****************************/
      MSGD(( INFO_XSER_PING_FROM, (char *)ConPktParamRec.szVersion ));
      ConPktParamRec.ulMessClass           = RSQL_PING_REPLY_EO003;
      memcpy(ConPktParamRec.szVersion, &en962_RTEVersionString[0], sizeof(ConPktParamRec.szVersion) );
#ifdef HAS_CONVERT_TO_PING_REPlY_VERSION_STRING
      sp100_ConvertToPingReplyVersionString(ConPktParamRec.szVersion);
#endif
      ConPktParamRec.ulCommState           = commErrOk_esp01;
      break;

    /****************************/
    default :
    /****************************/
      MSGD (( ERR_XSER_ILL_FRST_MESS_CLASS, ConPktParamRec.ulMessClass ));
      ulCommState                = commErrNotOk_esp01;
      ConPktParamRec.ulCommState = commErrNotOk_esp01;
      break;
    }

  if ( ulCommState != commErrOk_esp01 )            // - do not overwrite 'ulCommState'
    eo420SendConnectPacket ( pConnInfo->NiHandle, (SEND_PKT_FUNC_PTR_EO003)eo40NiSend,
                             &ConPktParamRec,  pErrText );
  else
    {
    ulCommState = eo420SendConnectPacket(  pConnInfo->NiHandle, (SEND_PKT_FUNC_PTR_EO003)eo40NiSend,
                                          &ConPktParamRec, pErrText );

    if ( ulCommState != commErrOk_esp01 )
      MSGD (( ERR_XSER_SEND_REQ_PACKET_ERR ));
    }

  return ( ulCommState );
}

//****************************************************************************
static teo00_Longuint
en962_NiTransfer (ten96_TCPIPConnectInfo *pConnInfo)
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_NiTransfer"));

  teo00_Longuint                  ulCommState = commErrOk_esp01;
  tsp00_ErrTextc                  pErrText;
  teo00_Longuint                  ulResDataLen;
  teo003_CommPacketRecord        *pResCommPacket;
  teo00_Longuint                  ulDataLen;
  teo00_Longuint                  ulSenderRef;
  teo00_Longuint                  ulReceiverRef;
  teo00_Longuint                  ulMessClass;

  do
  {
    ulCommState = eo420ReceiveCommPacket ( pConnInfo->NiHandle,
                                           (RECV_PKT_FUNC_PTR_EO003)eo40NiReceive,
                                           pConnInfo->ulSwapType,
                                           pConnInfo->pCommPacket,
                                           pConnInfo->ulMaxCmdDataLen,
                                           &ulDataLen,
                                           &ulSenderRef,
                                           &ulReceiverRef,
                                           &ulMessClass,
                                           pErrText );

    DBG1 ((FUNCTION_NAME_MEO60,"CommPacket recv. ( %d bytes ), Ref:%d, ulCommState = %d",
                  ulDataLen, pConnInfo->ulServerRef, ulCommState ));

    if ( ulCommState != commErrOk_esp01 )
    {
      if ( ulCommState == commErrReleased_esp01 )
      {
        ulCommState = commErrOk_esp01;
        break;
      }

      MSGD (( ERR_XSER_RECV_PACKET, pConnInfo->ulServerRef, (char *)pErrText ));
      break;
    }
    else if ( ! ulDataLen )
      break;
    else if ( ulMessClass != RSQL_USER_DATA_REQUEST_EO003 )
    {
      MSGD ((ERR_XSER_ILL_MESS_CLASS, ulMessClass, pConnInfo->ulServerRef));
      break;
    }

    DBG1 ((FUNCTION_NAME_MEO60,"Request packet read ( %d bytes ), Ref:%d",
                  ulDataLen, pConnInfo->ulServerRef ));

    if ( pConnInfo->pDBM != NULL )
    {
      ulCommState = eo908_Request( pConnInfo->pDBM, ulDataLen );

      if ( ulCommState != commErrOk_esp01 )
      {
        if ( ulCommState == commErrReleased_esp01 )
          ulCommState = commErrOk_esp01;
        break;
      }
      else if ( ! ulDataLen )
        break;

      DBG1 ((FUNCTION_NAME_MEO60,"Local request returned, Ref:%d, CommState = %d",
             pConnInfo->ulServerRef, ulCommState ));

      tsp00_Uint4         ResDataLen;


      ulCommState = eo908_Receive( pConnInfo->pDBM, (void**)&pResCommPacket,
                                   &ResDataLen );

      ulResDataLen = ResDataLen;


      if ( ulCommState != commErrOk_esp01 )
      {
        MSGD (( ERR_XSER_RECEIVE, pConnInfo->ulServerRef, (char *)pErrText ));
        break;
      }
    }
    else if ( pConnInfo->ulServiceType == srvPing_esp01 )
    {
        SAPDB_UInt1     hops;

        if ( eo420ExtractPingPacket( (char*)pConnInfo->pCommPacket->pDataPart, ulDataLen, &hops, 0, 0 ) )
        {
            if ( hops-- == 0 )
            {   // No more hops!
                pResCommPacket  = (teo003_CommPacketRecord*)pConnInfo->pCommPacket;

                ulResDataLen =  eo420FillPingPacket ( (char*) pResCommPacket->pDataPart,
                                                      pConnInfo->ulMaxCmdDataLen - ulDataLen +
                                                      pConnInfo->ulMinReplySize,
                                                      0, // hops equal to 0 means that the final destination was not reached
                                                      (char*)"XServer",
                                                      en962_RTEVersionString, 0 );

                ulCommState = eo420SendCommPacket ( pConnInfo->NiHandle, (SEND_PKT_FUNC_PTR_EO003)eo40NiSend,
                                                    pConnInfo->ulMaxSegmentSize,
                                                    pResCommPacket,
                                                    ulResDataLen,
                                                    ulSenderRef,
                                                    ulReceiverRef,
                                                    RSQL_USER_DATA_REPLY_EO003,
                                                    commErrOk_esp01,
                                                    pErrText );

                if ( ulCommState != commErrOk_esp01 )
                {
                    MSGD (( ERR_XSER_SEND_PACKET, pConnInfo->ulServerRef, (char *)pErrText ));
                    break;
                }
                continue;
            }

            ulCommState = en33Request ( &pConnInfo->Shm, ulDataLen, pErrText ) ;

            if ( ulCommState != commErrOk_esp01 )
            {
                MSGD (( ERR_XSER_REQUEST, pConnInfo->ulServerRef, (char *)pErrText ));
                break;
            }

            DBG1 ((FUNCTION_NAME_MEO60,"Local request returned, Ref:%d, CommState = %d",
                  pConnInfo->ulServerRef, ulCommState ));

            ulCommState = en33Receive( &pConnInfo->Shm, &pResCommPacket,
                                       &ulResDataLen  , pErrText );

            if ( ulCommState != commErrOk_esp01 )
                break;
        }
        else
        {
            ulCommState = commErrNotOk_esp01;
            break;
        }
    }
    else
    {
      ulCommState = en33Request ( &pConnInfo->Shm, ulDataLen, pErrText ) ;

      if ( ulCommState != commErrOk_esp01 )
      {
        MSGD (( ERR_XSER_REQUEST, pConnInfo->ulServerRef, (char *)pErrText ));
        break;
      }

      DBG1 ((FUNCTION_NAME_MEO60,"Local request returned, Ref:%d, CommState = %d",
             pConnInfo->ulServerRef, ulCommState ));

      ulCommState = en33Receive( &pConnInfo->Shm, &pResCommPacket,
                                 &ulResDataLen  , pErrText );

      if ( ulCommState != commErrOk_esp01 )
        break;
    }

    DBG1 ((FUNCTION_NAME_MEO60,"Local receive returned, Ref:%d, CommState = %d",
            pConnInfo->ulServerRef, ulCommState ));

    ulCommState = eo420SendCommPacket ( pConnInfo->NiHandle, (SEND_PKT_FUNC_PTR_EO003)eo40NiSend,
                                        pConnInfo->ulMaxSegmentSize,
                                        pResCommPacket,
                                        ulResDataLen,
                                        ulSenderRef,
                                        ulReceiverRef,
                                        RSQL_USER_DATA_REPLY_EO003,
                                        commErrOk_esp01,
                                        pErrText );
    if ( ulCommState != commErrOk_esp01 )
    {
      MSGD (( ERR_XSER_SEND_PACKET, pConnInfo->ulServerRef, (char *)pErrText ));
      break;
    }

    DBG1 ((FUNCTION_NAME_MEO60, "Reply packet send ( %d bytes ), Ref:%d",
            ulResDataLen, pConnInfo->ulServerRef ));
  }
  while (ulCommState == commErrOk_esp01) ;

  return ( ulCommState );
}

//****************************************************************************
static  teo00_Bool en962_AnalyzeParamAndSetGlobals (int              argc ,
                                                    char           **argv ,
                                                    teo00_Bool      *killServer,
                                                    teo00_Bool      *unlistenServer,
                                                    tsp01_RteError  *RteError)

//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_AnalyzeParamAndSetGlobals"));

  teo00_Bool    Ok = true;
  int argn, rc = 0;


  DBGIN;

  *killServer = false ;
  *unlistenServer = false ;

  for ( argn = 1 ; Ok && argn < argc ; argn ++ )
    {
	    if ( argv [ argn ] [ 0 ] != '-' ) break ;
	    switch ( argv [ argn ] [ 1 ] )
	    {
  	    case 'S' :
  		    if ( argv [ argn ] [ 2 ] )
            {
                en962_NiServicePort = (unsigned short )atoi(argv[argn] + 2 );
            }
  		    else
            {
                if ( argv[argn+1] )
  			        en962_NiServicePort = (unsigned short )atoi(argv[++argn] );
  			    else
  			        Ok = false ;
  		    }
  		    break ;
	    case 'i' :
		    e96_interactive = 1 ;
		    break ;
	    case 'k' :
            *killServer = true ;
             break ;
        case 'u':
            *unlistenServer = true;
            break;

	    case 'V' :
             printf ("%s (niserver)\n", (char *)en962_RTEVersionString );
             Ok = false ;
             break ;

        default :
		    MSGD (( ERR_NI_SERV_OPTION , argv [ argn ] [ 1 ] ))
		    Ok = false ;
		    break ;
	    }
    }

  return ( Ok);
}

//****************************************************************************
static  void en962_CloseHandles ()
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_CloseHandles"));

  /* redirect stdin to the null device */
  close ( 0 );
  open ( "/dev/null", O_RDONLY , 0 );

  /*
   *  Close all files which might possibly be open.
   */
  for ( int fd = sysconf ( _SC_OPEN_MAX ) - 1; fd > 2; --fd )
    close ( fd );

}
//****************************************************************************
static  teo00_Bool en962_InitNi (unsigned short  *NiServicePort,
                                 void           **NiListenHandle,
                                 tsp01_RteError  *RteError )
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_InitNi"));

  tsp00_Pathc               NiTraceFile ;

  if ( sql43_get_service_by_name ( (char *)SOCKET_SERVICE_NI72_EO003 , NiServicePort ) != 0 )
  {
    MSGD (( ERR_SERV_BY_NAME_ERROR , SOCKET_SERVICE_NI72_EO003 ))
	/* if the DB-installation installs the NI-service in "etc/services" */
#ifdef NI_SERVICE_INSTALLED
    eo46_set_rte_error ( RteError, undef_ceo00, "Service not found in /etc/services", NULL ) ;
    return false;
#endif
  }

  if ( BuildNiServerFilename( ni_tracefile, NiTraceFile, RteError ) )
  {
    tsp00_ErrTextc pErrText;
    /* remove old version */
    unlink ( NiTraceFile );

    if ( eo40NiInit(NiTraceFile,  NI_TRACE_ERRORS_ONLY_EO40, pErrText) == commErrOk_esp01 )
    {
        chmod(NiTraceFile, 0660);
        if ( eo40NiListen ( NiListenHandle, NiServicePort, pErrText) != commErrOk_esp01 )
        eo46_set_rte_error ( RteError, undef_ceo00, pErrText, NULL ) ;
      DBG1 ((FUNCTION_NAME_MEO60,"NI listening on port: %d", *NiServicePort ));
    }
    else
      eo46_set_rte_error ( RteError, undef_ceo00, pErrText, NULL ) ;
  }

  return(RteError->RteErrCode == RTE_NO_ERROR_SP01);
}

//****************************************************************************
static void en962_SetProtocolFile ( tsp01_RteError  *RteError )
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_SetProtocolFile"));

  static   tsp00_Pathc      ServerProtFile;

  if ( sqlGetIndependentProtPath ( ServerProtFile, TERM_WITH_DELIMITER_EO01, RteError ) )
  {
    if ( chdir ( ServerProtFile ) == UNDEF )
    {
      eo46_set_rte_error ( RteError, undef_ceo00, "change dir", ServerProtFile.asCharp() ) ;
      MSGD (( ERR_NI_SERV_CHDIR , ServerProtFile.asCharp() , RteError->OsErrText ))
    }
    else
    {
      if ( BuildNiServerFilename( ni_protfile, ServerProtFile, RteError ) )
      {
          unlink ( ServerProtFile.asCharp() );

          close ( 1 );
          close ( 2 );
          sql22_file = ServerProtFile.asCharp() ;
          sql22_timestamp = 1 ;
          sql22_fd = -1;
          sql22_keep_open = 0;
          MSGD (( INFO_NI_SERV_VERSION , (char *)en962_RTEVersionString ));
          MSGD (( INFO_NI_SERV_STARTED ));

          /* allow only owner to write to this file */
          chmod(ServerProtFile.asCharp(), 0660);
      }
    }
  }

}


//****************************************************************************
static  teo00_Bool en962_SetPidAndProcessGroup (pid_t           *MyPid ,
                                                pid_t           *MasterPid ,
                                                tsp01_RteError  *RteError )
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en962_SetPidAndProcessGroup"));

  teo00_Bool    Ok = true;

  *MyPid = getpid ();

#if defined(OSF1)
#ifdef setpgrp
#undef setpgrp
#endif	
#define setpgrp _Esetpgrp
#endif /* OSF1 */


  if ( setpgid (0, 0) != 0 )
  {
    eo46_set_rte_error ( RteError, undef_ceo00, "setpgrp ", NULL ) ;
    DBG1 ((FUNCTION_NAME_MEO60,": setpgrp failed, %s\n" , RteError->OsErrText )) ;
    DBG1 ((FUNCTION_NAME_MEO60,":            master_pid       %ld \n",	(long) *MasterPid )) ;
    DBG1 ((FUNCTION_NAME_MEO60,":            mupid  %ld \n",	(long) *MyPid )) ;
  }
  else
  {
    *MasterPid = *MyPid;
 	DBG1 ((FUNCTION_NAME_MEO60,"set process group %ld \n", (long) *MasterPid ));
  }

  return(RteError->RteErrCode == RTE_NO_ERROR_SP01);
}

/*==========================================================================*/

//****************************************************************************
static  void    en960_SetStdSignals ( int IsInteractive )
//****************************************************************************
{
  ROUTINE_DBG_MEO60 (_T("en960_SetStdSignals"));

  if ( IsInteractive )
  {
    (void) e96_SetSigAction ( SIGHUP, SIG_DFL );
    (void) e96_SetSigAction ( SIGINT, SIG_DFL );
  }
  else
  {
    (void) e96_SetSigAction ( SIGHUP, SIG_IGN );
    (void) e96_SetSigAction ( SIGINT, SIG_IGN );
  }

  (void) e96_SetSigAction ( SIGALRM , (SignalHandler)e96_catch_signal );
  (void) e96_SetSigAction ( SIGTERM , (SignalHandler)e96_catch_signal );
  (void) e96_SetSigAction ( SIGPIPE , SIG_IGN          );
  (void) e96_SetSigAction ( SIGCLD  , (SignalHandler)e96_catch_sigcld );
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/

//****************************************************************************
static  SIGNALFUNCTYPE  e96_catch_signal ( int			sig )
//****************************************************************************
{

  ROUTINE_DBG_MEO60 (_T("e96_catch_signal"));

  int				rc ;


    DBG1 ((FUNCTION_NAME_MEO60,"caught signal %d \n", sig ));

    switch ( sig )
    {
    case SIGHUP :
    case SIGINT :
    case SIGTERM :
		    MSGD (( INFO_NI_SERV_STOPPED ))
		    if ( (e96_pid == e96_master_pid) && (e96_master_pid > 1) )
		    {
                rc = kill ( (pid_t) (- e96_master_pid) , SIGTERM );
			    if ( rc == -1 )
			        DBG1 ((FUNCTION_NAME_MEO60,"group kill error, %s\n", "sqlerrs()" ));
		    }
//    (void) shutdown ( e96_cip->ci_sd , 2 );
//    (void) close ( e96_cip->ci_sd );
//    (void) close ( e96_sd );
		    exit ( sig );
		    break ;
    case SIGALRM :
            if ( GlobalConnnectInfo )
    	      GlobalConnnectInfo->Shm.ci_state = CON_TIMEDOUT ;
		    break ;
    }

}

/*---------------------------------------------------------------------------*/

static teo00_Longuint	en962InfoRequestToDB ( ten96_TCPIPConnectInfo   *pConnInfo,
                                             teo003_ConPktParamRecord *pConPktParamRec,
                                             tsp00_ErrTextc           pErrText )
{
    ROUTINE_DBG_MEO60 (_T("en962InfoRequestToDB"));

    int								rc ;
    int								kernel_fd ;
    int								reply_fd ;
    uid_t							uid ;
    tsp00_VFilename   file ;
    struct sockaddr   sa ;

    teo003_ConPktParamRecord ConPktParamRec;

    char         ucPacket[sizeof(teo003_RteHeaderRecord) +
                          sizeof(teo003_RteConnectPacketRec)];
    teo003_RteHeaderRecord     *pRTEHeader     = (teo003_RteHeaderRecord *)&ucPacket[0];
    teo003_RteConnectPacketRec *pConnectPacket = (teo003_RteConnectPacketRec *) &ucPacket[sizeof(teo003_RteHeaderRecord)];

    rte_header		*header = (rte_header *)&ucPacket[0];

    DBG1 ((FUNCTION_NAME_MEO60,"local dbname '%s' service %d \n",
          (char *)pConnInfo->Shm.ci_peer_dbname , pConnInfo->Shm.ci_service ));

    rc = sql32_open_kernel_fifo ( pConnInfo->Shm.ci_peer_dbname , &kernel_fd ,
                                  &uid , pErrText );
    if ( rc != commErrOk_esp01 )
    {
   	  DBG1 ((FUNCTION_NAME_MEO60,"cannot open knl fifo: %s\n", (char *)errmsg ));
      return ( rc );
    }
    DBG1 ((FUNCTION_NAME_MEO60,"opened kernel fifo, fd %d \n", kernel_fd ));

    /*
    *  The reply fifo must be open before REQUESTOR writes to it.
    *  Otherwise REQUESTOR might be fast enough to close the reply fifo
    *  before the application had it open.
    *  The last close on a fifo clears its contents.
    */
    rc = sql32_open_reply_fifo ( pConnInfo->Shm.ci_peer_dbname ,
                                 pConnInfo->WorkerPid ,
                                 (tsp00_Int4) 1 ,
                                 (char *)file ,
                                 &reply_fd ,
                                 pErrText );
    if ( rc != commErrOk_esp01 )
    {
      DBG1 ((FUNCTION_NAME_MEO60,"cannot open reply fifo: %s\n", (char *)pErrText ));
      (void) close ( kernel_fd );
      return ( rc );
    }

    /*
    *  An INFO_REQUEST can be delivered as received from application.
    *  Node name and reply fifo name must be added.
    */
    strcpy(pConPktParamRec->szOptPeerNode, (char *)pConnInfo->Shm.ci_peer_node );
    strcpy(pConPktParamRec->szOptFiFo    ,(char *)file );

    rc = eo420SendConnectPacket( (void *) &kernel_fd,
                                 (SEND_PKT_FUNC_PTR_EO003)en42SocketSendPacket,
                                 pConPktParamRec, pErrText );

    (void) close ( kernel_fd );

    if ( rc != commErrOk_esp01 )
    {
      DBG1 ((FUNCTION_NAME_MEO60,"cannot write to kernel fifo: %s \n", (char *)pErrText ));
      (void) close ( reply_fd );
      (void) unlink ( file );
      return ( rc );
    }

    DBG1 ((FUNCTION_NAME_MEO60,"closed fd %d \n", kernel_fd ));
    DBG1 ((FUNCTION_NAME_MEO60,"sent  info   request to '%s' \n", (char *)pConnInfo->Shm.ci_peer_dbname ));

    rc = eo420ReceiveConnectPacket ( (void *) &reply_fd,
                                     (RECV_PKT_FUNC_PTR_EO003)en42Receive,
                                     pConPktParamRec, pErrText );

    (void) close ( reply_fd );
    (void) unlink ( file );

    if ( rc != commErrOk_esp01 )
    {
      DBG1 ((FUNCTION_NAME_MEO60,": %s \n", (char *)pErrText ));
      return ( rc );
    }

    DBG1 ((FUNCTION_NAME_MEO60,"received reply from fd %d \n", reply_fd ));

    if ( pConPktParamRec->ulMessClass != RSQL_INFO_REPLY_EO003 )
    {
      DBG1 ((FUNCTION_NAME_MEO60,": got illegal reply packet, messclass %d \n",
             pConPktParamRec->ulMessClass ));
      return ( commErrNotOk_esp01 );
    }

    strcpy (pConPktParamRec->szReceiverServerDB, pConPktParamRec->szSenderServerDB ) ;

    DBG1 ((FUNCTION_NAME_MEO60,"MaxSegmentSize %d \n", pConPktParamRec->ulMaxSegmentSize ));
    DBG1 ((FUNCTION_NAME_MEO60,"MaxDataLen     %d \n", pConPktParamRec->ulMaxDataLen ));
    DBG1 ((FUNCTION_NAME_MEO60,"ulPacketSize   %d \n", pConPktParamRec->ulPacketSize ));
    DBG1 ((FUNCTION_NAME_MEO60,"ulMinReplySize %d \n", pConPktParamRec->ulMinReplySize ));
    DBG1 ((FUNCTION_NAME_MEO60,"returning %d \n", rc ));
    return ( rc );
}

//****************************************************************************
static  SIGNALFUNCTYPE  e96_catch_sigcld ( int             sig)
//****************************************************************************
{
    ROUTINE_DBG_MEO60 (_T("e96_catch_signal"));

    (void) wait ((int *) 0);
}

//****************************************************************************
#ifdef ALARM
static  SIGNALFUNCTYPE  e96_catch_alarm (
int             sig)
//****************************************************************************
{
    ROUTINE_DBG_MEO60 (_T("e96_catch_alarm"));

    int				    state ;
    tsp00_Int4			    srvref ;
    tsp00_Int4			    cliref ;
    pid_t			    clipid ;
    struct comseg_header	    *comseg ;
    struct communication_segment    *ocomseg ;
    char			    buffer [ 256 ];


    DBG1 ((FUNCTION_NAME_MEO60,"old_db %d \n", e96_old_db ));
    if ( e96_old_db )
    {
	ocomseg = sql03_condat->oc_comseg ;
	DBG1 ((FUNCTION_NAME_MEO60,"ocomseg 0x%08lx \n", (long) ocomseg )) ;
	if ( ! ocomseg )
	{
	    MSGD (( INFO_NI_SERV_DROPPED ))
	    exit ( 2 );
	}
	state  = ocomseg->cs_state ;
	srvref = ocomseg->cs_ukp_reference ;
	cliref = ocomseg->cs_apc_reference ;
	clipid = ocomseg->cs_apc_pid ;
	DBG1 ((FUNCTION_NAME_MEO60,"state %d sref %d(%d) cref %d cpid %ld \n",
	       state , (int) srvref , (int) sql03_condat->oc_ukp_reference ,
	       (int) cliref , (long) clipid )) ;

	if ( (clipid != e96_pid) ||
	     (cliref != e96_cip->ci_my_ref) )
	{
	    DBG1 ((FUNCTION_NAME_MEO60,"assuming timeout \n" )) ;
	    state = SP1CE_TIMEOUT ; /* comseg given away, assuming timeout */
	}
	if ( (state != SP1CE_OK) ||
	     (srvref != sql03_condat->oc_ukp_reference) )
	{
	    MSGD (( INFO_NI_SERV_DROPPED ))
		sql42_send_control ( e96_cip->ci_sd ,
				     RSQL_USER_RELEASE_REQUEST_EO003 ,
				     e96_cip->ci_peer_ref , e96_remote_ref ,
				     state );
	    DBG1 ((FUNCTION_NAME_MEO60,"waiting for application to close \n" )) ;
	    while ( recv ( e96_cip->ci_sd , buffer , sizeof(buffer) , 0 ) > 0 )
	    {
		/* consume packets until EOF */
	    }
	    DBG1 ((FUNCTION_NAME_MEO60,"closing and exiting \n" )) ;
	    (void) shutdown ( e96_cip->ci_sd , 2 );
	    (void) close ( e96_cip->ci_sd );
	    exit ( 2 );
	}
    }
    else
    {
	comseg = e96_cip->ci_comseg ;
	DBG1 ((FUNCTION_NAME_MEO60,"comseg 0x%08lx \n", (long) comseg )) ;
	if ( ! comseg )
	{
	    MSGD (( INFO_NI_SERV_DROPPED ))
	    exit ( 2 );
	}
	state = comseg->cs_server_state ;
	srvref = comseg->cs_server_ref ;
	cliref = comseg->cs_client_ref ;
	clipid = comseg->cs_client_pid ;
	DBG1 ((FUNCTION_NAME_MEO60,"state %d sref %d(%d) cref %d cpid %ld \n",
	       state , (int) srvref , (int) e96_cip->ci_peer_ref ,
	       (int) cliref , (long) clipid )) ;

	if ( (clipid != e96_pid) ||
	     (cliref != e96_cip->ci_my_ref) )
	{
	    DBG1 ((FUNCTION_NAME_MEO60,"assuming timeout \n" )) ;
	    state = SP1CE_TIMEOUT ; /* comseg given away, assuming timeout */
	}
	if ( (state != SP1CE_OK) ||
	     (srvref != e96_cip->ci_peer_ref) )
	{
	    MSGD (( INFO_NI_SERV_DROPPED ))
		sql42_send_control ( e96_cip->ci_sd ,
				     RSQL_USER_RELEASE_REQUEST_EO003 ,
				     e96_cip->ci_peer_ref , e96_remote_ref ,
				     state );
	    DBG1 ((FUNCTION_NAME_MEO60,"waiting for application to close \n" )) ;
	    while ( recv ( e96_cip->ci_sd , buffer , sizeof(buffer) , 0 ) > 0 )
	    {
		/* consume packets until EOF */
	    }
	    DBG1 ((FUNCTION_NAME_MEO60,"closing and exiting \n" )) ;
	    (void) shutdown ( e96_cip->ci_sd , 2 );
	    (void) close ( e96_cip->ci_sd );
	    exit ( 2 );
	}
    }

}
#endif

/*------------------------------*/

static int e96_SetSigAction(int sig, SignalHandler Handler )
{
  struct sigaction sa;

  sa.sa_handler = Handler;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0;
  return sigaction( sig, &sa, NULL);
}
/*==========================================================================*/

